/*
 *  linux/fs/cse451/dir.c
 *
 *  Copyright (C) 1991, 1992 Linus Torvalds
 *
 *  cse451 directory handling functions
 */

#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/stat.h>
#include <linux/fcntl.h>
#include <linux/quotaops.h>
#include <linux/locks.h>

#include <asm/uaccess.h>

#include "../cse451fs.h"

#define PARENT_INO(buffer) \
(((struct cse451_dir_entry *) ((buffer)+CSE451_DIRENTRYSIZE))->inode)


//////////////////////////////////////////////////////////////////
// cse451-only helper functions                                 //
//////////////////////////////////////////////////////////////////

static inline int namecompare(int len, int maxlen,
	const char * name, const char * buffer)
{
	if (len < maxlen && buffer[len])
		return 0;
	return !memcmp(name, buffer, len);
}

/*
 *	cse451_add_entry()
 *
 * adds a file entry to the specified directory, returning a possible
 * error value if it fails.
 *
 * NOTE!! The inode part of 'de' is left at 0 - which means you
 * may not sleep between calling this and putting something into
 * the entry, as someone else might have used it while you slept.
 */
static int cse451_add_entry(struct inode * dir,
	const char * name, int namelen,
	struct buffer_head ** res_buf,
	struct cse451_dir_entry ** res_dir)
{
	int i;
	unsigned long block, offset;
	struct buffer_head * bh;
	struct cse451_dir_entry * de;

	DEBUGPRINTK (("_add_entry\n"))
	DEBUGPRINTK (("... '%13s' -> inode %ld\n", name, dir->i_ino))

	*res_buf = NULL;
	*res_dir = NULL;

	if (namelen > CSE451_MAXDIRNAMELENGTH) {
		return -ENAMETOOLONG;
	}
	if (!namelen)
		return -ENOENT;

	bh = NULL;
	block = offset = 0;

	while (1) {
		if (!bh) {
			bh = cse451_bread(dir,block,1);
			if (!bh)
				return -ENOSPC;
		}
		de = (struct cse451_dir_entry *) (bh->b_data + offset);
		offset += CSE451_DIRENTRYSIZE;
		if (block*bh->b_size + offset > dir->i_size) {
			de->inode = 0;
			dir->i_size = block*bh->b_size + offset;
			mark_inode_dirty(dir);
		}
		if (!de->inode) {
			dir->i_mtime = dir->i_ctime = CURRENT_TIME;
			mark_inode_dirty(dir);
			for (i = 0; i < CSE451_MAXDIRNAMELENGTH; i++)
				de->name[i] = (i < namelen) ? name[i] : 0;
			dir->i_version = ++event;
			mark_buffer_dirty(bh);
			*res_dir = de;
			break;
		}
		if (offset < bh->b_size)
			continue;
		brelse(bh);
		bh = NULL;
		offset = 0;
		block++;
	}
	*res_buf = bh;
	return 0;
}

/*
 *	cse451_find_entry()
 *
 * finds an entry in the specified directory with the wanted name. It
 * returns the cache buffer in which the entry was found, and the entry
 * itself (as a parameter - res_dir). It does NOT read the inode of the
 * entry - you'll have to do that yourself if you want to.
 */
static struct buffer_head * cse451_find_entry(struct inode * dir,
	const char * name, int namelen, struct cse451_dir_entry ** res_dir)
{
	unsigned long block, offset;
	struct buffer_head * bh;
	struct cse451_sb_info * pSBInfo;
	struct cse451_dir_entry *de;

	*res_dir = NULL;
	pSBInfo = (struct cse451_sb_info*)dir->i_sb->u.generic_sbp;

	if (namelen > CSE451_MAXDIRNAMELENGTH) namelen = CSE451_MAXDIRNAMELENGTH;

	bh = NULL;
	block = offset = 0;
	while (block*BLOCK_SIZE+offset < dir->i_size) {
		if (!bh) {
			bh = cse451_bread(dir,block,0);
			if (!bh) {
				block++;
				continue;
			}
		}
		de = (struct cse451_dir_entry *) (bh->b_data + offset);
		offset += CSE451_DIRENTRYSIZE;
		if (de->inode && namecompare(namelen, CSE451_MAXDIRNAMELENGTH,name,de->name)) {
			*res_dir = de;
			return bh;
		}

		if (offset >= bh->b_size) {
			brelse(bh);
			bh = NULL;
			offset = 0;
			block++;
		}
	}

	brelse(bh);

	return NULL;
}

/*
 * routine to check that the specified directory is empty (for rmdir)
 */
static int empty_dir(struct inode * inode)
{
	unsigned int block, offset;
	struct buffer_head * bh;
	struct cse451_dir_entry * de;

	DEBUGPRINTK (("empty_dir\n"))

	block = 0;
	bh = NULL;
	offset = 2*CSE451_DIRENTRYSIZE;
	if (inode->i_size & (CSE451_DIRENTRYSIZE-1))
		goto bad_dir;
	if (inode->i_size < offset)
		goto bad_dir;
	bh = cse451_bread(inode,0,0);
	if (!bh)
		goto bad_dir;
	de = (struct cse451_dir_entry *) bh->b_data;
	if (!de->inode || strcmp(de->name,"."))
		goto bad_dir;
	de = (struct cse451_dir_entry *) (bh->b_data + CSE451_DIRENTRYSIZE);
	if (!de->inode || strcmp(de->name,".."))
		goto bad_dir;
	while (block*BLOCK_SIZE+offset < inode->i_size) {
		if (!bh) {
			bh = cse451_bread(inode,block,0);
			if (!bh) {
				block++;
				continue;
			}
		}
		de = (struct cse451_dir_entry *) (bh->b_data + offset);
		offset += CSE451_DIRENTRYSIZE;
		if (de->inode) {
			brelse(bh);
			return 0;
		}
		if (offset < bh->b_size)
			continue;
		brelse(bh);
		bh = NULL;
		offset = 0;
		block++;
	}
	brelse(bh);
	return 1;
bad_dir:
	brelse(bh);
	printk("Bad directory on device %s\n",
	       kdevname(inode->i_dev));
	return 1;
}


//////////////////////////////////////////////////////////////////
// VFS callbacks                                                //
//////////////////////////////////////////////////////////////////

// create a file (it's not even on disk at the moment)
int cse451_create(struct inode * dir, struct dentry *dentry, int mode)
{
	int error;
	struct inode * inode;
	struct buffer_head * bh;
	struct cse451_dir_entry * de;

	DEBUGPRINTK (("_create: '%s' mode = 0x%08x\n", dentry->d_name.name, mode))

	inode = cse451_new_inode(dir, &error);
	if (!inode) return error;

	inode->i_op = &cse451_file_inode_operations;
	inode->i_fop = &cse451_file_operations;
	inode->i_mapping->a_ops = &cse451_aops;
	inode->i_mode = mode;
	mark_inode_dirty(inode);

	error = cse451_add_entry(dir, dentry->d_name.name,
				dentry->d_name.len, &bh ,&de);
	if (error) {
		inode->i_nlink--;
		mark_inode_dirty(inode);
		iput(inode);
		return error;
	}

	de->inode = inode->i_ino;

	mark_buffer_dirty(bh);
	brelse(bh);

	d_instantiate(dentry, inode);

	DEBUGPRINTK (("cse451_create: created '%s' at inode %d (addr 0x%p)\n", dentry->d_name.name, inode->i_ino, ((struct cse451_inode_info*)inode->u.generic_ip)->i_datablocks[0]))

	return 0;
}


// 
// VFS is looking for the file whose name is in 'dentry' in the directory
// described by 'dir'.
// 
struct dentry *cse451_lookup(struct inode * dir, struct dentry *dentry)
{
	struct inode * inode = NULL;
	struct cse451_dir_entry * de;
	struct buffer_head * bh;

	DEBUGPRINTK (("_lookup: %ld '%13s'\n", dentry->d_inode?dentry->d_inode->i_ino:-1, dentry->d_name.name))

	dentry->d_op = dir->i_sb->s_root->d_op;

	bh = cse451_find_entry(dir, dentry->d_name.name, dentry->d_name.len, &de);
	if (bh) {
		brelse (bh);
		inode = iget(dir->i_sb, de->inode);
 		if (!inode) return ERR_PTR(-EACCES);
	}
	d_add(dentry, inode);

	return NULL;
}


//
// Create a hard link
//
int cse451_link(struct dentry * old_dentry, struct inode * dir,
	       struct dentry *dentry)
{
	int error;
	struct inode* 		 inode;
	struct cse451_dir_entry* de;
	struct buffer_head*      bh;

	DEBUGPRINTK (("_link\n"))

	inode = old_dentry->d_inode;

	if (S_ISDIR(inode->i_mode)) return -EPERM;
	if (inode->i_nlink >= CSE451_LINK_MAX) return -EMLINK;

	error = cse451_add_entry(dir, dentry->d_name.name,
				dentry->d_name.len, &bh, &de);
	if (error) {
		brelse(bh);
		return error;
	}

	de->inode = inode->i_ino;

	mark_buffer_dirty(bh);
	brelse(bh);

	inode->i_nlink++;
	//inode->i_ctime = CURRENT_TIME;

	mark_inode_dirty(inode);
	atomic_inc(&inode->i_count);

	d_instantiate(dentry, inode);

	return 0;
}

int cse451_unlink(struct inode * dir, struct dentry *dentry)
{
	int retval;
	struct inode * inode;
	struct buffer_head * bh;
	struct cse451_dir_entry * de;

	DEBUGPRINTK (("_unlink\n"))
	DEBUGPRINTK (("... '%13s'\n", dentry->d_name.name))

	retval = -ENOENT;
	inode = dentry->d_inode;

	bh = cse451_find_entry(dir, dentry->d_name.name,
				dentry->d_name.len, &de);

	if (!bh || de->inode != inode->i_ino) {
		DEBUGPRINTK (("unlink problem: 0x%p:%d - %d\n", bh, (int)de->inode, (int)inode->i_ino))

		goto end_unlink;
	}

	if (!inode->i_nlink) {
		printk("Deleting nonexistent file (%s:%lu), %d\n",
			kdevname(inode->i_dev),
		       inode->i_ino, inode->i_nlink);
		inode->i_nlink=1;
	}

	de->inode = 0;	// mark the dentry in the file as "unused"

	dir->i_version = ++event;  // used to indicate the directory has
				   // changed (odd kind of race avoidance)

	mark_buffer_dirty(bh);	 // cause directory block to be written back
	mark_inode_dirty(dir);	 // schedule

	inode->i_nlink--;	 // update the link count
	mark_inode_dirty(inode); // new link count needs to be saved

	DEBUGPRINTK (("_unlink: inode nlink/count = %d %d\n", (int)inode->i_nlink, (int)atomic_read(&inode->i_count)))

	//iput(inode);

	retval = 0;

end_unlink:
	brelse(bh);
	return retval;
}

int cse451_mkdir(struct inode * dir, struct dentry *dentry, int mode)
{
	int error;
	struct inode * inode;
	struct buffer_head * bh, *dir_block;
	struct cse451_dir_entry * de;

	DEBUGPRINTK (("_mkdir\n"))
	if (dentry) {
		DEBUGPRINTK (("... %d '%13s'\n", atomic_read(&dentry->d_count),dentry->d_name.name))
	}

	if (dir->i_nlink >= CSE451_LINK_MAX) return -EMLINK;

	inode = cse451_new_inode(dir, &error);
	if (!inode) return error;

	inode->i_op = &cse451_dir_inode_operations;
	inode->i_fop = &cse451_dir_operations;
	inode->i_size = 2 * CSE451_DIRENTRYSIZE;
	inode->i_blocks = 0;

	dir_block = cse451_bread(inode,0,1);
	if (!dir_block) {
		inode->i_nlink--;
		mark_inode_dirty(inode);
		iput(inode);
		return -ENOSPC;
	}

	de = (struct cse451_dir_entry *) dir_block->b_data;
	de->inode=inode->i_ino;
	strcpy(de->name,".");
	de = (struct cse451_dir_entry *) (dir_block->b_data + 
							CSE451_DIRENTRYSIZE);
	de->inode = dir->i_ino;
	strcpy(de->name,"..");

	inode->i_nlink = 2;
	mark_buffer_dirty(dir_block);
	brelse(dir_block);
	inode->i_mode = S_IFDIR | mode;
	if (dir->i_mode & S_ISGID)
		inode->i_mode |= S_ISGID;
	mark_inode_dirty(inode);

	error = cse451_add_entry(dir, dentry->d_name.name,
				dentry->d_name.len, &bh, &de);
	DEBUGPRINTK (("_mkdir: add_entry -> %d\n",error))
	if (error) {
		inode->i_nlink=0;
		iput(inode);
		return error;
	}

	de->inode = inode->i_ino;

	dir->i_nlink++;
	mark_inode_dirty(dir);

	mark_buffer_dirty(bh);

	d_instantiate(dentry, inode);

	DEBUGPRINTK (("_mkdir done %d %d 0x%p\n", dentry->d_inode->i_ino, atomic_read(&dentry->d_count), dentry))

	brelse(bh);

	return 0;
}

int cse451_rmdir(struct inode * dir, struct dentry *dentry)
{
	int retval;
	struct inode * inode;
	struct buffer_head * bh;
	struct cse451_dir_entry * de;

	DEBUGPRINTK (("rm_dir\n"))

	inode = NULL;
	bh = cse451_find_entry(dir, dentry->d_name.name,
			      dentry->d_name.len, &de);
	retval = -ENOENT;
	if (!bh)
		goto end_rmdir;
	inode = dentry->d_inode;

	if (!empty_dir(inode)) {
		retval = -ENOTEMPTY;
		goto end_rmdir;
	}
	if (de->inode != inode->i_ino) {
		retval = -ENOENT;
		goto end_rmdir;
	}
	if (inode->i_nlink != 2)
		printk("empty directory has nlink!=2 (%d)\n",inode->i_nlink);
	de->inode = 0;
	dir->i_version = ++event;
	mark_buffer_dirty(bh);
	inode->i_nlink=0;
	mark_inode_dirty(inode);
	inode->i_ctime = dir->i_ctime = dir->i_mtime = CURRENT_TIME;
	dir->i_nlink--;
	mark_inode_dirty(dir);
	retval = 0;
end_rmdir:
	brelse(bh);
	return retval;
}

int cse451_mknod(struct inode * dir, struct dentry *dentry, int mode, int rdev)
{
	int error;
	struct inode * inode;
	struct buffer_head * bh;
	struct cse451_dir_entry * de;

	DEBUGPRINTK (("_mknod\n"))
	DEBUGPRINTK (("... '%13s'\n", dentry->d_name.name))

	inode = cse451_new_inode(dir, &error);
	if (!inode) return error;

	inode->i_uid = current->fsuid;
	init_special_inode(inode, mode, rdev);

	mark_inode_dirty(inode);

	error = cse451_add_entry(dir, dentry->d_name.name,
					dentry->d_name.len, &bh, &de);
	if (error) {
		inode->i_nlink--;
		mark_inode_dirty(inode);
		iput(inode);
		return error;
	}

	de->inode = inode->i_ino;

	mark_buffer_dirty(bh);
	brelse(bh);

	d_instantiate(dentry, inode);

	return 0;
}


/*
 * Anybody can rename anything with this: the permission checks are left to
 * the higher-level routines.
 */

int cse451_rename(struct inode * old_dir, struct dentry *old_dentry,
			   struct inode * new_dir, struct dentry *new_dentry)
{
	struct inode * old_inode, * new_inode;
	struct buffer_head * old_bh, * new_bh, * dir_bh;
	struct cse451_dir_entry * old_de, * new_de;
	int retval;

	DEBUGPRINTK (("_rename\n"))
	DEBUGPRINTK (("... '%13s' -> '%13s'\n", old_dentry->d_name.name, new_dentry->d_name.name))

	new_bh = dir_bh = NULL;

	old_inode = old_dentry->d_inode;
	new_inode = new_dentry->d_inode;

	old_bh = cse451_find_entry(old_dir, old_dentry->d_name.name,
				  old_dentry->d_name.len, &old_de);

	retval = -ENOENT;
	if (!old_bh || old_de->inode != old_inode->i_ino)
		goto end_rename;

	retval = -EPERM;
	new_bh = cse451_find_entry(new_dir, new_dentry->d_name.name,
				  new_dentry->d_name.len, &new_de);
	if (new_bh) {
		if (!new_inode) {
			brelse(new_bh);
			new_bh = NULL;
		}
	}

	if (S_ISDIR(old_inode->i_mode)) {
		if (new_inode) {
			retval = -ENOTEMPTY;
			if (!empty_dir(new_inode))
				goto end_rename;
		}
		retval = -EIO;
		dir_bh = cse451_bread(old_inode,0,0);
		if (!dir_bh) goto end_rename;

		if (PARENT_INO(dir_bh->b_data) != old_dir->i_ino)
			goto end_rename;

		retval = -EMLINK;
		if (!new_inode && new_dir != old_dir &&
				new_dir->i_nlink >=CSE451_LINK_MAX)
			goto end_rename;
	}

	if (!new_bh) {
		retval = cse451_add_entry(new_dir,
					 new_dentry->d_name.name,
					 new_dentry->d_name.len,
					 &new_bh, &new_de);
		if (retval)
			goto end_rename;
	}

	/* ok, that's it */

	new_de->inode = old_inode->i_ino;
	old_de->inode = 0;

	old_dir->i_ctime = old_dir->i_mtime = CURRENT_TIME;

	old_dir->i_version = ++event;

	mark_inode_dirty(old_dir);

	new_dir->i_ctime = new_dir->i_mtime = CURRENT_TIME;
	new_dir->i_version = ++event;
	mark_inode_dirty(new_dir);

	if (new_inode) {
		new_inode->i_nlink--;
		new_inode->i_ctime = CURRENT_TIME;
		mark_inode_dirty(new_inode);
	}

	mark_buffer_dirty(old_bh);
	mark_buffer_dirty(new_bh);

	if (dir_bh) {
		PARENT_INO(dir_bh->b_data) = new_dir->i_ino;
		mark_buffer_dirty(dir_bh);
		old_dir->i_nlink--;
		mark_inode_dirty(old_dir);
		if (new_inode) {
			new_inode->i_nlink--;
			mark_inode_dirty(new_inode);
		} else {
			new_dir->i_nlink++;
			mark_inode_dirty(new_dir);
		}
	}
	retval = 0;

end_rename:
	brelse(dir_bh);
	brelse(old_bh);
	brelse(new_bh);

	return retval;
}

// supports readdir system call

int cse451_readdir(struct file * filp, void * dirent, filldir_t filldir)
{
	unsigned int			offset;
	struct buffer_head* 		bh;
	struct cse451_dir_entry*	de;
	struct inode*			inode;

	DEBUGPRINTK (("_readdir\n"))


	inode = filp->f_dentry->d_inode;

 	// file position better be aligned to a directory entry...
	if (filp->f_pos & (CSE451_DIRENTRYSIZE - 1)) return -EBADF;

	while (filp->f_pos < inode->i_size) {

		offset = filp->f_pos & (BLOCK_SIZE-1);

		bh = cse451_bread(inode,(filp->f_pos)>>BLOCK_SIZE_BITS,0);
		if (!bh) {
			filp->f_pos += BLOCK_SIZE-offset;
			continue;
		}

		do {
			de = (struct cse451_dir_entry *)(offset + bh->b_data);
			// skip unused dirent's
			if (de->inode) {
				int size = strnlen(de->name, CSE451_MAXDIRNAMELENGTH);
				if (filldir(dirent, de->name, size, filp->f_pos, de->inode, DT_UNKNOWN) < 0) {
					brelse(bh);
					return 0;
				}
			}

			offset += CSE451_DIRENTRYSIZE;
			filp->f_pos += CSE451_DIRENTRYSIZE;

		} while (offset < BLOCK_SIZE && filp->f_pos < inode->i_size);

		brelse(bh);
	}

	return 0;
}

//---------------------------------------------------------


struct file_operations cse451_dir_operations = {
	read:		generic_read_dir,
	readdir:	cse451_readdir,
	fsync:		file_fsync,
};

