/*
 *  linux/fs/cse451/inode.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  Copyright (C) 1996  Gertjan van Wingerde    (gertjan@cs.vu.nl)
 *	Minix V2 fs support.
 *
 *  Modified for 680x0 by Andreas Schwab
 */

#include <linux/module.h>

#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/locks.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <linux/highuid.h>
#include <linux/blkdev.h>

#include <asm/system.h>
#include <asm/bitops.h>

#include "../cse451fs.h"

static struct super_operations cse451_sops = {
        read_inode:     cse451_read_inode,
        write_inode:    cse451_write_inode,
        delete_inode:   cse451_delete_inode,
	clear_inode:	cse451_clear_inode,
        put_super:      cse451_put_super,
        write_super:    cse451_write_super,
        statfs:         cse451_statfs,
};

//////////////////////////////////////////////////////////////////
// cse451-only helper functions                                 //
//////////////////////////////////////////////////////////////////

int cse451_new_block(struct inode * inode)
{
	struct super_block* 	sb;
	struct cse451_sb_info*	pSBInfo;
	struct buffer_head* 	bh;
	int			mapBlock, bitOffset = 0, blockNum;

	sb = inode->i_sb;

	if (!sb) {
		printk("trying to get new block from nonexistent device\n");
		return 0;
	}
	pSBInfo = (struct cse451_sb_info*)sb->u.generic_sbp;

repeat:
	bh = NULL;
	for (mapBlock = 0; mapBlock<pSBInfo->pSB->s_nDataBlocks; mapBlock++) {
		bh = pSBInfo->pDmap[mapBlock];
		if (!bh) break;
		for (bitOffset=0; bitOffset<BLOCK_SIZE*8; bitOffset++) {
			if (!test_bit(bitOffset, bh->b_data)) break;
		}
		if (bitOffset < BLOCK_SIZE*8) break;
	}
	if (!bh || bitOffset >= BLOCK_SIZE*8) return 0;

	if (test_and_set_bit(bitOffset,bh->b_data)) {
		DEBUGPRINTK (("new_block: bit already set"))
		goto repeat;
	}

	blockNum = mapBlock*8192 + bitOffset + pSBInfo->pSB->s_nDataBlocksStart;

	mark_buffer_dirty(bh);

	return blockNum;
}

void cse451_free_block(struct inode * inode, int block)
{
        struct super_block * sb;
        struct buffer_head * bh;
        unsigned int bit, zone;
	unsigned int dataStart;
	struct cse451_sb_info*	pSBInfo;

	sb = inode->i_sb;	
        if (!sb) {
                printk("trying to free block on nonexistent device\n");
                return;
        }

	pSBInfo = (struct cse451_sb_info*)sb->u.generic_sbp;
	dataStart = pSBInfo->pSB->s_nDataBlocksStart;

        if (block < dataStart ||
            block >=dataStart + pSBInfo->pSB->s_nDataBlocks) {
                printk("trying to free block not in datazone: %d\n", block);
                return;
        }

        bh = get_hash_table(sb->s_dev,block,BLOCK_SIZE);

        if (bh)
                clear_bit(BH_Dirty, &bh->b_state);
        brelse(bh);

        zone = block - dataStart;
        bit = zone & (8*BLOCK_SIZE - 1);
        zone >>= (BLOCK_SIZE_BITS+3);	// divide by BLOCK_SIZE * 8
        if (zone >= pSBInfo->pSB->s_nDataBlocks) {
                printk("cse451_free_block: nonexistent bitmap buffer\n");
                return;
        }

        bh = pSBInfo->pDmap[zone];

	DEBUGPRINTK (("free_block: block %d dataStart %d zone %d bit %d\n",block,(int)dataStart,(int)zone,(int)bit))

        if (!test_and_clear_bit(bit,bh->b_data))
                printk("cse451_free_block (%s:%d): bit already cleared\n",
                       kdevname(sb->s_dev), block);

        mark_buffer_dirty(bh);

        return;
}

int get_block(struct inode * inode, long block,
                        struct buffer_head *bh_result, int create)
{
        int err = -EIO;
	struct cse451_inode_info* pInodeInfo;

	if (block<0) {
		printk ("cse451fs: get_block( <0 )\n");
		return err;
	}
	if (block >= CSE451_NUMDATAPTRS) {
		printk ("cse451fs: get_block( >%d )\n", CSE451_NUMDATAPTRS);
		return err;
	}

	pInodeInfo = (struct cse451_inode_info*)inode->u.generic_ip;

	if (pInodeInfo->i_datablocks[block] == 0) {
		if (!create) return -EFAULT;
		pInodeInfo->i_datablocks[block] = cse451_new_block(inode);
		if (pInodeInfo->i_datablocks[block]==0) return -ENOSPC;
		bh_result->b_state |= (1UL << BH_New);
	}


        bh_result->b_dev = inode->i_dev;
        bh_result->b_blocknr = pInodeInfo->i_datablocks[block];
        bh_result->b_state |= (1UL << BH_Mapped);

        return 0;

}

struct buffer_head *cse451_getblk(struct inode *inode, int block, int create)
{
        struct buffer_head dummy;
        int error;

        dummy.b_state = 0;
        dummy.b_blocknr = -1000;
        error = get_block(inode, block, &dummy, create);
        if (!error && buffer_mapped(&dummy)) {
                struct buffer_head *bh;
                bh = getblk(dummy.b_dev, dummy.b_blocknr, BLOCK_SIZE);
                if (buffer_new(&dummy)) {
                        memset(bh->b_data, 0, BLOCK_SIZE);
                        mark_buffer_uptodate(bh, 1);
                        mark_buffer_dirty(bh);
                }
                return bh;
        }
        return NULL;
}

struct buffer_head * cse451_bread(struct inode * inode, int block, int create)
{
        struct buffer_head * bh;

        bh = cse451_getblk(inode, block, create);
        if (!bh || buffer_uptodate(bh))
                return bh;
        ll_rw_block(READ, 1, &bh);
        wait_on_buffer(bh);
        if (buffer_uptodate(bh))
                return bh;
        brelse(bh);
        return NULL;
}

static int nibblemap[] = { 4,3,3,2,3,2,2,1,3,2,2,1,2,1,1,0 };

// how many free data blocks are there?
// (The last block of the map may have only partial information,
// but we rely on the "phantom blocks" (off the end of the real map)
// to be marked busy.)

static unsigned long count_free(struct buffer_head *map[],
				 unsigned numblocks, __u32 numbits)
{
        unsigned i, j, sum = 0;
        struct buffer_head *bh;

        for (i=0; i<numblocks; i++) {
                if (!(bh=map[i])) {
			printk ("cse451fs: unexpected error - map %d is null\n", i);
                        return 0;
		}
                for (j=0; j<BLOCK_SIZE; j++)
                        sum += nibblemap[bh->b_data[j] & 0xf]
                                + nibblemap[(bh->b_data[j]>>4) & 0xf];
        }

        return(sum);
}

//////////////////////////////////////////////////////////////////
// VFS callbacks                                                //
//////////////////////////////////////////////////////////////////

void cse451_put_super(struct super_block *sb)
{
	int i;
	struct cse451_sb_info*	pSBInfo;

	DEBUGPRINTK (("_put_super\n"))

	pSBInfo = (struct cse451_sb_info*)sb->u.generic_sbp;

	for (i = 0; i < pSBInfo->pSB->s_nDataMapBlocks; i++) {
		brelse(pSBInfo->pDmap[i]);
	}

	brelse (pSBInfo->sb_bh);

	kfree(pSBInfo->pDmap);
	kfree(pSBInfo);
	sb->u.generic_sbp = NULL;

	return;
}

void cse451_write_super(struct super_block * sb)
{
	struct cse451_sb_info*	pSBInfo;

	DEBUGPRINTK (("_write_super\n"))

	pSBInfo = (struct cse451_sb_info*)sb->u.generic_sbp;
	mark_buffer_dirty(pSBInfo->sb_bh);

	sb->s_dirt = 0;	// indicate that we've arranged for it to be written
}

struct super_block *cse451_read_super(struct super_block *s,
						void *data, int silent)
{

	struct buffer_head*        bh;
	struct buffer_head**       map;
	struct cse451_super_block* pSB;
	struct cse451_sb_info*	   pSBInfo;
	struct inode*              root_inode;
	int i, block = 0;
	unsigned int hblock;
	kdev_t dev = s->s_dev;

	DEBUGPRINTK (("_read_super\n"))
	
	hblock = get_hardsect_size(dev);
	if (hblock && hblock > BLOCK_SIZE) {
		printk("cse451fs: fs blocksize too small for device.\n");
		return NULL;
	}

	s->s_blocksize = BLOCK_SIZE;
	s->s_blocksize_bits = BLOCK_SIZE_BITS;
	s->s_op = &cse451_sops;
	set_blocksize(dev, BLOCK_SIZE);

	DUMPX4 (s->s_blocksize);
	DUMPX1 (s->s_blocksize_bits);

	// get the cse451fs superblock

	if (!(bh = bread(dev,1,BLOCK_SIZE))) {
		printk("cse451fs: unable to read superblock\n");
		return NULL;
	}

	pSB = (struct cse451_super_block *) bh->b_data;

	s->s_magic = pSB->s_magic;
	if (s->s_magic != CSE451_SUPER_MAGIC) {
		printk ("cse451fs: bad superblock magic number: 0x%x\n", pSB->s_magic);
		goto out_release;
	}

	// alloc the sb_info structure

	pSBInfo = kmalloc(sizeof(struct cse451_sb_info), GFP_KERNEL);
	if (!pSBInfo) goto out_nomem;

	s->u.generic_sbp = pSBInfo;

	pSBInfo->sb_bh = bh;
	pSBInfo->pSB = pSB;
	pSBInfo->pImap = pSB->s_imap;

	// set root directory
	root_inode = iget(s, CSE451_ROOT_INO);
	if (!root_inode) {
		if (!silent) printk("cse451fs: get root inode failed\n");
		goto out_free_sb_info;
	}

	s->s_root = d_alloc_root(root_inode);
	if (!s->s_root) {
		printk ("cse451fs: couldn't d_alloc_root()\n");
		iput(root_inode);
		goto out_freemap;
	}

	s->s_root->d_op = NULL;

        DUMPX1 (s->s_dirt);
        DEBUGPRINTK (("file_system_type:name = %s\n", s->s_type->name))
	DUMPX4 (s->s_type->fs_flags);
        DUMPX4 (s->s_flags);
        DUMPX4 (s->s_root);
        DUMPX4 (s->s_magic);

	map = kmalloc(pSB->s_nDataMapBlocks * sizeof(bh), GFP_KERNEL);
	if (!map) {
		printk ("cse451fs: Can't allocate mem for data map\n");
		goto out_free_sb_info;
	}
	pSBInfo->pDmap = map;

	memset(map, 0, pSB->s_nDataMapBlocks*sizeof(bh));

	for (block=0 ; block < pSB->s_nDataMapBlocks; block++) {
		if (!(pSBInfo->pDmap[block]=bread(dev,block+2,BLOCK_SIZE))) {
			printk("cse451fs: bad superblock or unable to read data map\n");
			goto out_freemap;
		}
	}

	return s;

out_nomem:
	printk("cse451fs: out of kernel memory\n");
	goto out_release;

out_freemap:
	for (i = 0; i < block; i++)
		brelse(pSBInfo->pDmap[i]);
	kfree(pSBInfo->pDmap);

out_free_sb_info:
	kfree( s->u.generic_sbp );

out_release:
	brelse(bh);
	return NULL;

}

int cse451_statfs(struct super_block *sb, struct statfs *buf)
{
	struct cse451_sb_info*  pSBInfo;

	DEBUGPRINTK (("_statfs\n"))

	pSBInfo = (struct cse451_sb_info*)sb->u.generic_sbp;

	buf->f_type = sb->s_magic;
	buf->f_bsize = sb->s_blocksize;
	buf->f_blocks = pSBInfo->pSB->s_nDataBlocks;

	buf->f_bfree = count_free(pSBInfo->pDmap, pSBInfo->pSB->s_nDataMapBlocks, pSBInfo->pSB->s_nDataBlocks);

	buf->f_bavail = buf->f_bfree;
	buf->f_files = pSBInfo->pSB->s_nBusyInodes;

	buf->f_ffree = pSBInfo->pSB->s_nNumInodes - pSBInfo->pSB->s_nBusyInodes;
	buf->f_namelen = CSE451_MAXDIRNAMELENGTH;
	return 0;
}

//////////////////////////////////////////////////////////////////
// module bootstrap registration                                //
//////////////////////////////////////////////////////////////////

static DECLARE_FSTYPE_DEV(cse451_fs_type,"cse451fs",cse451_read_super);

static int __init init_cse451_fs(void)
{
        return register_filesystem(&cse451_fs_type);
}

static void __exit exit_cse451_fs(void)
{
        unregister_filesystem(&cse451_fs_type);
}

EXPORT_NO_SYMBOLS;

module_init(init_cse451_fs)
module_exit(exit_cse451_fs)
