/*
 * mkfs.myfs.c - make a cse451 file system
 *
 */

#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <stdlib.h>
#include <termios.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <mntent.h>
#include <getopt.h>

#include <linux/fs.h>
#include "../cse451fs.h"

#include "nls.h"

#ifndef __GNUC__
#error "needs gcc for the bitop-__asm__'s"
#endif

#ifndef __linux__
#define volatile
#endif

#include "bitops.h"

#define UPPER(size,n)      ((size+((n)-1))/(n))

// inode name '0' is not usable, so we adjust for that here
#define mark_inode_inuse(x) (setbit(Super->s_imap,(x)-1))
#define mark_inode_free(x) (clrbit(Super->s_imap,(x)-1))
#define inode_in_use(x) (bit(inode_map,(x)-1))

// datablock '0' means "not allocated" (to allow holes in files)
#define mark_datablk_inuse(x) (setbit(datablk_map,(x)-1))
#define mark_datablk_free(x) (clrbit(datablk_map,(x)-1))
#define datablk_in_use(x) (bit(datablk_map,(x)-1))

static char * program_name = "mkfs";
static char boot_block_buffer[BLOCK_SIZE];
static char *datablk_map;

volatile void die(char *str) {
	fprintf(stderr, "%s: %s\n", program_name, str);
	exit(8);
}

volatile void usage()
{
	fprintf(stderr, "%s (%s)\n", program_name, util_linux_version);
	fprintf(stderr,
		_("Usage: %s [-c | -l filename] [-nXX] [-iXX] /dev/name [blocks]\n"),
		  program_name);
	exit(16);
}

/*
 * Check to make certain that our new filesystem won't be created on
 * an already mounted partition.  Code adapted from mke2fs, Copyright
 * (C) 1994 Theodore Ts'o.  Also licensed under GPL.
 */
static void check_mount(const char* devname)
{
	FILE * f;
	struct mntent * mnt;

	if ((f = setmntent (MOUNTED, "r")) == NULL) return;

	while ((mnt = getmntent (f)) != NULL)
		if (strcmp (devname, mnt->mnt_fsname) == 0) break;
	endmntent (f);

	if (!mnt) return;

	die(_("%s is mounted; will not make a filesystem here!"));
}

static long valid_offset (int fd, int offset)
{
	char ch;

	if (lseek (fd, offset, 0) < 0)
		return 0;
	if (read (fd, &ch, 1) < 1)
		return 0;
	return 1;
}

static int count_blocks (int fd)
{
	int high, low;

	low = 0;
	for (high = 1; valid_offset (fd, high); high *= 2)
		low = high;
	while (low < high - 1)
	{
		const int mid = (low + high) / 2;

		if (valid_offset (fd, mid))
			low = mid;
		else
			high = mid;
	}
	valid_offset (fd, 0);
	return (low + 1);
}

static int get_size(const char  *file)
{
	int	fd;
	long	size;

	fd = open(file, O_RDWR);
	if (fd < 0) {
		perror(file);
		exit(1);
	}
	if (ioctl(fd, BLKGETSIZE, &size) >= 0) {
		close(fd);
		return (size * 512);
	}

	size = count_blocks(fd);
	close(fd);
	return size;
}

void write_data(int DEV, int offset, void* buffer, int len)
{
	if (offset != lseek(DEV, offset, SEEK_SET))
		die(_("seek failed in write_data"));
	if (len != write(DEV, buffer, len))
		die(_("write failed in write_data"));
}

void write_tables(int DEV, struct cse451_super_block* Super)
{
	write_data(DEV, 0, boot_block_buffer, BLOCK_SIZE);
	write_data(DEV, BLOCK_SIZE, (char*)Super, BLOCK_SIZE);

	write_data(DEV, BLOCK_SIZE*2, datablk_map, Super->s_nDataMapBlocks*BLOCK_SIZE);

}

void make_root_dir(int DEV, struct cse451_super_block* Super)
{
	struct cse451_inode inode;
	char root_block[BLOCK_SIZE] = "\0";
	char* tmp;
	int block;

	mark_inode_inuse(CSE451_ROOT_INO);

	// fill root ("/") directory
	tmp = root_block;
	*(short *)tmp = CSE451_ROOT_INO;
	strcpy(tmp+2,".");

	tmp += CSE451_DIRENTRYSIZE;
	*(short *)tmp = CSE451_ROOT_INO;
	strcpy(tmp+2,"..");

	tmp += CSE451_DIRENTRYSIZE;
	*tmp = '\0';
	*(tmp+1) = '\0';

	write_data(DEV, BLOCK_SIZE* Super->s_nDataBlocksStart, root_block, BLOCK_SIZE);
	mark_datablk_inuse(1);

	inode.i_mode = S_IFDIR + 0755;
	inode.i_nlinks = 2;   // I point at myself twice ("." and "..")
	inode.i_filesize = 2*CSE451_DIRENTRYSIZE;

	inode.i_datablocks[0] = Super->s_nDataBlocksStart;
	for (block=1; block<CSE451_NUMDATAPTRS; block++) {
		inode.i_datablocks[block] = 0;	// mark as unalloc'ed
	}

	write_data(DEV, Super->s_nInodeStart*BLOCK_SIZE, &inode, sizeof(inode));
}

void setup_tables(const char* devname, struct cse451_super_block* Super)
{
	int i;
	unsigned long totalBlocks;
	unsigned long numInodes;
	unsigned long numInodeBlocks;
	unsigned long numDataBlocks;
	unsigned long numDataMapBlocks;

	memset(boot_block_buffer,0,BLOCK_SIZE);

	memset(Super,0,BLOCK_SIZE);
	Super->s_magic = CSE451_SUPER_MAGIC;

 	totalBlocks = get_size (devname) / BLOCK_SIZE;
	if (totalBlocks<10) {
		die("Device is too small");
	}

	// inode map is whatever is left of superblock's block...
	numInodes = (BLOCK_SIZE - SUPERBLOCK_HEADERLENGTH)*8;
	if (numInodes > totalBlocks / 3) {
		// just a heuristic - make sure there's room for data!
		numInodes = totalBlocks / 3;
	}
	numInodeBlocks = (numInodes+CSE451_INODES_PER_BLOCK-1)
				& ~(CSE451_INODES_PER_BLOCK-1);
	numInodeBlocks /= CSE451_INODES_PER_BLOCK;

	// crude bounding analysis of allocation of remaining
	// blocks between data and data map
	totalBlocks = totalBlocks - 2 - numInodeBlocks;
	numDataMapBlocks = totalBlocks / (BLOCK_SIZE*8) + 1;
	numDataBlocks = totalBlocks - numDataMapBlocks;

	Super->s_nNumInodes = numInodes;
	Super->s_nBusyInodes = 1;	// the root inode
	Super->s_nDataMapStart = 2;
	Super->s_nDataMapBlocks = numDataMapBlocks;
	Super->s_nInodeStart = 2 + numDataMapBlocks;
	Super->s_nNumInodeBlocks = numInodeBlocks;
	Super->s_nDataBlocksStart = 2 + numDataMapBlocks + numInodeBlocks;
	Super->s_nDataBlocks = numDataBlocks;

	printf ("Decided on %d inodes\n", numInodes);
	printf ("%d blocks devoted to inodes\n", Super->s_nNumInodeBlocks);
	printf ("%d blocks devoted to data map\n", Super->s_nDataMapBlocks);
	printf ("%d blocks devoted to data\n", Super->s_nDataBlocks);

	memset(Super->s_imap, 0xff, BLOCK_SIZE-SUPERBLOCK_HEADERLENGTH);
	for (i = CSE451_ROOT_INO ; i<=Super->s_nNumInodes ; i++)
		mark_inode_free(i);

	// 1 block for boot, 1 for superblock, plus blocks for inodes.
	// rest can be used for data blocks and data block map

	datablk_map = malloc(numDataMapBlocks * BLOCK_SIZE);
	if (!datablk_map) die(_("unable to allocate buffers for maps"));

	memset(datablk_map, 0xff, numDataMapBlocks * BLOCK_SIZE);
	for (i = 1 ; i<=Super->s_nDataBlocks ; i++) {
		mark_datablk_free(i);
	}

}

int main(int argc, char ** argv)
{
	int    i;
	char * tmp;
	struct stat statbuf;
	char * p;
	char * device_name = NULL;
	int    DEV;

	char super_block_buffer[BLOCK_SIZE];
	struct cse451_super_block* Super;

	Super = (struct cse451_super_block*)super_block_buffer;

	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);

	if (argc && *argv) program_name = *argv;
	if ((p = strrchr(program_name, '/')) != NULL) program_name = p+1;

	if (CSE451_INODE_SIZE * CSE451_INODES_PER_BLOCK != BLOCK_SIZE)
		die(_("bad inode size - doesn't divide block size"));

	if (argc > 1 && !device_name) {
		device_name = argv[1];
	} else {
		usage();
	}

	check_mount(device_name);  /* is it already mounted? */

	DEV = open(device_name,O_RDWR );
	if (DEV<0) die(_("unable to open %s"));
	if (fstat(DEV,&statbuf)<0) die(_("unable to stat %s"));

	setup_tables(device_name, Super);

	make_root_dir(DEV, Super);

	write_tables(DEV, Super);

	return 0;
}
